/*
 * Support for flash modules on ATI Xilleon 220 platforms
 *
 * Copyright (C) Robert Lembree <lembree@metrolink.com>
 *
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <asm/io.h>
// #include <asm/arch/hardware.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <asm/ati/xilleon.h>


__u8 xilleon_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb(map->map_priv_1 + ofs);
}

__u16 xilleon_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw(map->map_priv_1 + ofs);
}

__u32 xilleon_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl(map->map_priv_1 + ofs);
}

void xilleon_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, map->map_priv_1 + adr);
	//printk("writeb %x at %x\n", d, map->map_priv_1 + adr);
	mb();
	asm __volatile__ ("sync");
}

void xilleon_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	//printk("writew %x at %x\n", d, map->map_priv_1 + adr);
	mb();
	asm __volatile__ ("sync");
}

void xilleon_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, map->map_priv_1 + adr);
	//printk("writel %x at %x\n", d, map->map_priv_1 + adr);
	mb();
	asm __volatile__ ("sync");
}

void xilleon_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	// printk ("xilleon_copy_from: 0x%x@0x%x -> 0x%x\n", len, from, to);
	memcpy_fromio(to, map->map_priv_1 + from, len);
	asm __volatile__ ("sync");
}

void xilleon_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	while(len) {
		__raw_writeb(*(unsigned char *) from, map->map_priv_1 + to);
		asm __volatile__ ("sync");
		from++;
		to++;
		len--;
	}
}


static struct mtd_info *core_flash_mtd;

struct map_info xilleon_core_flash_map = {
	name: "Core flash",
	size: XILLEON_CORE_FLASH_SIZE,
	buswidth: 1,
	read8: xilleon_read8,
	read16: xilleon_read16,
	read32: xilleon_read32,
	copy_from: xilleon_copy_from,
	write8: xilleon_write8,
	write16: xilleon_write16,
	write32: xilleon_write32,
	copy_to: xilleon_copy_to
};



static int __init init_xilleon_core_flash (void)
{
	xilleon_core_flash_map.map_priv_1 = 
		(unsigned long)ioremap(XILLEON_CORE_FLASH_BASE, 
				       XILLEON_CORE_FLASH_SIZE);
	core_flash_mtd = do_map_probe("jedec_probe", &xilleon_core_flash_map);
	if (!core_flash_mtd) {
		printk("Boot FLASH probe failed\n");
		return -ENXIO;
	}

	core_flash_mtd->module = THIS_MODULE;
	
	if (add_mtd_device(core_flash_mtd)) {
		printk("Boot FLASH device addition failed\n");
		map_destroy(core_flash_mtd);
		core_flash_mtd = 0;
		return -ENOMEM;
	}
		
	return 0;
}




static struct mtd_info *user_flash_mtd;

struct map_info xilleon_user_flash_map = {
	name: "User flash",
	size: XILLEON_USER_FLASH_SIZE,
	buswidth: 2,
	read8: xilleon_read8,
	read16: xilleon_read16,
	read32: xilleon_read32,
	copy_from: xilleon_copy_from,
	write8: xilleon_write8,
	write16: xilleon_write16,
	write32: xilleon_write32,
	copy_to: xilleon_copy_to
};



static int __init init_xilleon_user_flash (void)
{
	xilleon_user_flash_map.map_priv_1 = 
		(unsigned long)ioremap(XILLEON_USER_FLASH_BASE, 
				       XILLEON_USER_FLASH_SIZE);
	user_flash_mtd = do_map_probe("jedec_probe", &xilleon_user_flash_map);

	if (!user_flash_mtd) {
		printk("User FLASH probe failed\n");
		return -ENXIO;
	}

	user_flash_mtd->module = THIS_MODULE;
	
	if (add_mtd_device(user_flash_mtd)) {
		printk("User FLASH device addition failed\n");
		map_destroy(user_flash_mtd);
		user_flash_mtd = 0;
		return -ENOMEM;
	}
		
	return 0;
}



static int __init init_xilleon_maps(void)
{

       	printk(KERN_INFO "XILLEON Flash mappings: \n");

	init_xilleon_core_flash();
	init_xilleon_user_flash();
	
	return 0;
}
	

static void __exit cleanup_xilleon_maps(void)
{
	if (core_flash_mtd) {
		del_mtd_device(core_flash_mtd);
		map_destroy(core_flash_mtd);
	}


	if (user_flash_mtd) {
		del_mtd_device(user_flash_mtd);
		map_destroy(user_flash_mtd);
	}

}

module_init(init_xilleon_maps);
module_exit(cleanup_xilleon_maps);

